// fpclassify.hpp

#ifndef BOOST_MATH_FPCLASSIFY_HPP
#define BOOST_MATH_FPCLASSIFY_HPP

// Copyright (c) 2006 Johan Rade

// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt
// or copy at http://www.boost.org/LICENSE_1_0.txt)

#include <cmath>
#include <cstring>
#include <boost/cstdint.hpp>
#include <boost/detail/endian.hpp>

#ifdef BOOST_PDP_ENDIAN
#	error PDP-endian platforms are not supported
#endif

namespace boost {
namespace math {

//------------------------------------------------------------------------------

namespace detail {

// The function copy copies the four most significant bytes
// of a floating point number

// The number of different copy functions could be reduced
// by more use of templates, but that does hurt the performance slightly
// (at least with VC7.1).

//..............................................................................

#if defined BOOST_BIG_ENDIAN

template<class T> void copy(T x, boost::uint32_t& a )
{
    memcpy(&a, &x, 4);
}

template<class T> void copy(boost::uint32_t a, T& x)
{
    memcpy(&x, &a, 4);
}

//..............................................................................

#elif defined BOOST_LITTLE_ENDIAN

template<class T> void copy(T x, boost::uint32_t& a )
{
    memcpy(
        &a,
        reinterpret_cast<const unsigned char*>(&x) + sizeof(T) - 4,
        4
    );
}

template<class T> void copy(boost::uint32_t a, T& x)
{
    memcpy(
        reinterpret_cast<unsigned char*>(&x) + sizeof(T) - 4,
        &a,
        4
    );
}

// sizeof(long double) is 8, 12 or 16.
// It seems that on all little endian processors with sizeof(long double) > 8,
// only 10 bytes are significant.
// (There are big endian processors with 16 significant bytes.)
// On these platforms we replace sizeof(long double) - 4 by 10 - 4 = 6.

inline void copy(long double x, boost::uint32_t& a )
{
    memcpy(
        &a,
        reinterpret_cast<const unsigned char*>(&x)
            + (sizeof(long double) == 8 ? 4 : 6),
        4
    );
}

inline void copy(boost::uint32_t a, long double& x)
{
    memcpy(
        reinterpret_cast<unsigned char*>(&x)
            + (sizeof(long double) == 8 ? 4 : 6),
        &a,
        4
    );
}

//..............................................................................

#else
#	error We should not get here
#endif

//------------------------------------------------------------------------------

template<class T> struct mask
{
    static const boost::uint32_t sign;
	static const boost::uint32_t exponent;
	static const boost::uint32_t significand;
	static const boost::uint32_t fraction;
};

template<> struct mask<float>
{
	static const boost::uint32_t sign =        0x80000000;
	static const boost::uint32_t exponent =    0x7f800000;
	static const boost::uint32_t significand = 0x00000000; 
	static const boost::uint32_t fraction =    0x007fffff;
};

template<> struct mask<double>
{
	static const boost::uint32_t sign =        0x80000000;
	static const boost::uint32_t exponent =    0x7ff00000;
	static const boost::uint32_t significand = 0x00000000; 
	static const boost::uint32_t fraction =    0x000fffff;
};

template<> struct mask<long double>
{
	static const boost::uint32_t sign =        0x80000000;

	// defined in long_double_mask.cpp
	static const boost::uint32_t exponent;
	static const boost::uint32_t significand;
	static const boost::uint32_t fraction;
};

}   // namespace detail

//------------------------------------------------------------------------------

template<class T> bool (isfinite)(T x)
{
    boost::uint32_t a;
    detail::copy(x,a);
    a &= detail::mask<T>::exponent;
    return a != detail::mask<T>::exponent;
}

//..............................................................

template<class T> bool (isnormal)(T x)
{
	boost::uint32_t a;
	detail::copy(x,a);
	a &= detail::mask<T>::exponent;
	return (a != 0) && (a != detail::mask<T>::exponent);
}

//............................................................

template<class T> bool (isinf)(T x)
{
	boost::uint32_t a;
	detail::copy(x,a);
	if((a & detail::mask<T>::exponent) != detail::mask<T>::exponent)
		return false;
	a &= detail::mask<T>::fraction;
	detail::copy(a,x);
	return x == 0;
}

inline bool (isinf)(float x)
{
    boost::uint32_t a;
    detail::copy(x,a);
    a &= ~detail::mask<float>::sign;
    return a == detail::mask<float>::exponent;
}

//...........................................................

template<class T> bool (isnan)(T x)
{
	boost::uint32_t a;
	detail::copy(x,a);
	if((a & detail::mask<T>::exponent) != detail::mask<T>::exponent)
		return false;
	a &= detail::mask<T>::fraction;
	detail::copy(a,x);
	return x != 0;
}

inline bool (isnan)(float x)
{
	boost::uint32_t a;
	detail::copy(x,a);
	return (a & detail::mask<float>::exponent) == detail::mask<float>::exponent
		   && (a & detail::mask<float>::fraction) != 0;
}

//------------------------------------------------------------------------------

#ifndef FP_INFINITE
#   define FP_INFINITE 0
#   define FP_NAN 1
#   define FP_NORMAL 2
#   define FP_SUBNORMAL 3
#   define FP_ZERO 4
#endif

template<class T> int (fpclassify)(T x)
{
    boost::uint32_t a;
    detail::copy(x,a);
    a &= detail::mask<T>::exponent;

    if(a == 0) {
        if(x)
            return FP_SUBNORMAL;
        else
            return FP_ZERO;
    }

    else if(a == detail::mask<T>::exponent) {
        if((isinf)(x))								// This can be optimized
			return FP_INFINITE;
        else
			return FP_NAN;
    }

    else
        return FP_NORMAL;
}

//------------------------------------------------------------------------------

template<class T> bool (signbit)(T x)
{
    boost::uint32_t a;
    detail::copy(x,a);
    a &= detail::mask<T>::sign;
    return a != 0;
}

template<class T> T (copysign)(T x, T y)      // magnitude of x and sign of y
{
    boost::uint32_t a;
    detail::copy(x,a);
    a &= ~detail::mask<T>::sign;

    boost::uint32_t b;
    detail::copy(y,b);
    b &= detail::mask<T>::sign;

    detail::copy(a|b,x);
    return x;
}

template<class T> T (changesign)(T x)
{
    boost::uint32_t a;
    detail::copy(x,a);
    a ^= detail::mask<T>::sign;
    detail::copy(a,x);
    return x;
}

//------------------------------------------------------------------------------

}   // namespace math
}   // namespace boost

#endif

