// test.cpp

// Copyright (c) 2006 Johan Rade

// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt
// or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifdef _MSC_VER
#   pragma warning(disable: 4127 4244 4511 4512 4701 4800)
#endif

//-------------------------------------

#include "../fpclassify.hpp"

#undef NDEBUG

#include <cassert>
#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <limits>

#include <boost/cstdint.hpp>

//------------------------------------------------------------------------------

using boost::math::fpclassify;
using boost::math::isfinite;
using boost::math::isnormal;
using boost::math::isinf;
using boost::math::isnan;

using boost::math::signbit;
using boost::math::copysign;
using boost::math::changesign;

//------------------------------------------------------------------------------

template<class T> void test_basic();
template<class T> void test_copysign();
template<class T> void test_changesign();

//------------------------------------------------------------------------------

int main()
{
#if defined BOOST_BIG_ENDIAN
    std::cout << "big endian\n";
#elif defined BOOST_LITTLE_ENDIAN
    std::cout << "little endian\n";
#else
    std::cout << "We should never get here\n";
#endif

    std::cout << "sizeof(long double) = " << sizeof(long double) << "\n\n";

	if(sizeof(long double) != sizeof(double)) {
	
		std::cout << std::hex << std::right << std::setfill('0');
		std::cout << "mask<long double>::sign        = "
			<< std::setw(8) << boost::math::detail::mask<long double>::sign
			<< "\n";
		std::cout << "mask<long double>::exponent    = "
			<< std::setw(8) << boost::math::detail::mask<long double>::exponent
			<< "\n";
		std::cout << "mask<long double>::significand = "
			<< std::setw(8) << boost::math::detail::mask<long double>::significand
			<< "\n";
		std::cout << "mask<long double>::fraction    = "
			<< std::setw(8) << boost::math::detail::mask<long double>::fraction
			<< "\n\n";
	}

	std::cout << "Testing float" << std::endl;
    test_basic<float>();
	test_copysign<float>();
	test_changesign<float>();

	std::cout << "Testing double" << std::endl;
	test_basic<double>();
    test_copysign<double>();
	test_changesign<double>();

	std::cout << "Testing long double" << std::endl;
	test_basic<long double>();
    test_copysign<long double>();
    test_changesign<long double>();

	std::cout << std::endl;

    return 0;
}

//------------------------------------------------------------------------------

template<class T>
void test_basic()
{
    assert(std::numeric_limits<T>::has_infinity);
	assert(std::numeric_limits<T>::has_quiet_NaN);
	assert(std::numeric_limits<T>::has_signaling_NaN);
    T x;

    x = static_cast<T>(0);
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_ZERO);

    x = (changesign)(static_cast<T>(0));
    assert((signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_ZERO);

    x = std::numeric_limits<T>::denorm_min();
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_SUBNORMAL);

    x = -std::numeric_limits<T>::denorm_min();
    assert((signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_SUBNORMAL);
/*
    // fail on some platforms due to incorrect definition of 
    // std::numeric_limits<long double>::min()

    x = std::numeric_limits<T>::min() / 2;
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_SUBNORMAL);

    x = -std::numeric_limits<T>::min() / 2;
    assert((signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_SUBNORMAL);
*/
    x = (std::numeric_limits<T>::min)();
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = -(std::numeric_limits<T>::min)();
    assert((signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = static_cast<T>(1);
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = static_cast<T>(-1);
    assert((signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = (std::numeric_limits<T>::max)();
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = -(std::numeric_limits<T>::max)();
    assert((signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = std::numeric_limits<T>::infinity();
    assert(!(signbit)(x));
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert((isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_INFINITE);

    x = -std::numeric_limits<T>::infinity();
    assert((signbit)(x));
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert((isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_INFINITE);

    x = std::numeric_limits<T>::quiet_NaN();
    assert(!(signbit)(x));
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert((isnan)(x));
    assert((fpclassify)(x) == FP_NAN);

    x = (changesign)(std::numeric_limits<T>::quiet_NaN());
    assert((signbit)(x));
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert((isnan)(x));
    assert((fpclassify)(x) == FP_NAN);

    x = std::numeric_limits<T>::signaling_NaN();
    assert(!(signbit)(x));
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert((isnan)(x));
    assert((fpclassify)(x) == FP_NAN);

    x = (changesign)(std::numeric_limits<T>::signaling_NaN());
    assert((signbit)(x));
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert((isnan)(x));
    assert((fpclassify)(x) == FP_NAN);
}

//------------------------------------------------------------------------------

#define countof(A) (sizeof(A)/sizeof((A)[0]))

template<class T> void test_copysign_impl(T x, T y);

template<class T> void test_copysign()
{
    T a[] = {
        static_cast<T>(0),
        std::numeric_limits<T>::denorm_min(),
        (std::numeric_limits<T>::min)() / 2,
        (std::numeric_limits<T>::min)(),
        static_cast<T>(1),
        (std::numeric_limits<T>::max)(),
        std::numeric_limits<T>::infinity(),
    };

    for(int i = 0; i < countof(a); ++i) {
        for(int j = 0; j < countof(a); ++j)
            test_copysign_impl(a[i], a[j]);
    }

    for(int i = 0; i < countof(a); ++i)
        test_copysign_impl(a[i], std::numeric_limits<T>::quiet_NaN());
}

template<class T> void test_copysign_impl(T x, T y)
{
    assert((copysign)(x, y) == x);
    assert((copysign)(x, (changesign)(y)) == (changesign)(x));
    assert((copysign)((changesign)(x), y) == x);
    assert((copysign)((changesign)(x), (changesign)(y)) == (changesign)(x));
}

//------------------------------------------------------------------------------

template<class T> void test_changesign()
{
    T a[] = {
        std::numeric_limits<T>::denorm_min(),
        (std::numeric_limits<T>::min)(),
        static_cast<T>(1),
        std::numeric_limits<T>::infinity()
    };

    for(int i = 0; i < countof(a); ++i) {
        assert((changesign)(a[i]) == -a[i]);
        assert((signbit)((changesign)(a[i])));
        assert((changesign)(-a[i]) == a[i]);
        assert(!(signbit)((changesign)(-a[i])));
    }

    T b = static_cast<T>(0);
    assert((changesign)(b) == b);
    assert((signbit)((changesign)(b)));
    assert((changesign)((changesign)(b)) == b);
    assert(!(signbit)((changesign)((changesign)(b))));

    T c = std::numeric_limits<T>::quiet_NaN();
    assert((isnan)((changesign)(c)));
    assert((signbit)((changesign)(c)));
    assert((isnan)((changesign)((changesign)(c))));
    assert(!(signbit)((changesign)((changesign)(c))));
}

